/*
 ****************************************************************************
 *
 * Copyright (C) 2004-2007 Duke University.
 * Copyright (c) 2009 University of Massachusetts, Amherst.
 * 
 * This software is distributed under
 * the terms of the Eclipse Public License Version 1.0 found in
 * the file named LICENSE.Eclipse, which was shipped with this distribution.
 * Any use, reproduction or distribution of this software constitutes
 * the recipient's acceptance of the Eclipse license terms.
 * This notice and the full text of the license must be included with any
 * distribution of this software.
 * 
 ****************************************************************************
 *
 * This is the main loop of the DieselNet Controller for ORCA. It is
 * invoked every tick, where it looks to see if a request for a lease has
 * been queued by an XML-RPC request. It also has methods that are invoked
 * as a lease transitions states.
 */

package orca.controllers.dieselnet;

import orca.cod.Node;
import orca.cod.NodeGroup;
import orca.cod.api.ICodClientReservation;

//import orca.container.ContainerManager;
import orca.shirako.container.ContainerManager;

import orca.manage.container.Manager;

import orca.policy.core.BrokerSimplePolicy;
import orca.policy.core.ServiceManagerCalendarPolicy;
import orca.policy.core.util.PropertiesManager;
import orca.policy.core.util.ResourceProperties;

import orca.shirako.api.IActor;
import orca.shirako.api.IController;
import orca.shirako.api.IReservation;
import orca.shirako.api.IServiceManager;
import orca.shirako.api.ISlice;
import orca.shirako.common.ResourceType;
import orca.shirako.core.EventHandler;
import orca.shirako.kernel.CodServiceManagerReservationFactory;
import orca.shirako.kernel.ResourceSet;
import orca.shirako.time.Term;

import orca.tools.images.core.ImageDescriptor;

import orca.util.LoggingTool;
import orca.util.PropList;

import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.LinkedList;
import java.util.concurrent.locks.ReentrantReadWriteLock;


import org.apache.xmlrpc.webserver.WebServer;
import org.apache.xmlrpc.server.PropertyHandlerMapping;
import org.apache.xmlrpc.server.XmlRpcServerConfigImpl;
import org.apache.xmlrpc.server.XmlRpcServer;

public class DieselnetController implements IController, Runnable {
	
	/*
	 * The methods in this object get called as the ORCA ticket transitions
	 * through its states. Mainly for efficiency, we don't record all states.
	 */
	protected class DieselnetEventHandler extends EventHandler {
		private void updateIfFailed(IReservation r) {
			// Does this ever evaluate to true?
			if (r.isFailed()) {
				leaseUpdateState(r, Lease.FAILED);
			}
		}
		
		@Override
		public void onBeforeExtendTicket(IReservation r) {
			updateIfFailed(r);
		}
		
		@Override
		public void onClose(IReservation r) {
			// Set closed here to ensure we see it.
			leaseUpdateState(r, Lease.CLOSED);
	    }

		@Override
		public void onCloseComplete(IReservation r) {
			updateIfFailed(r);
		}
		
		@Override
		public void onExtendLease(IReservation r) {
			updateIfFailed(r);
		}
		
		@Override
		public void onExtendLeaseComplete(IReservation r) {
			// We don't currently support extending leases. Update
			// status start and end times here if/when support is
			// added.
			updateIfFailed(r);
	    }
		
		@Override
	    public void onExtendTicket(IReservation r) {
			updateIfFailed(r);
	    }

		@Override
	    public void onExtendTicketComplete(IReservation r) {
			updateIfFailed(r);
	    }
		
		@Override
		public void onLease(IReservation r) {
			updateIfFailed(r);
		}

		@Override
		public void onLeaseComplete(IReservation r) {
			Lease lease = leaseUpdateState(r, Lease.LEASED);
			// We make note of the start and end times. These are simply
			// for use by the getLeases() status request; we do not take
			// any action based on these values.
			leasesLock.writeLock().lock();
			try {
				Term term = r.getApprovedTerm();
				lease.actStart = term.getStartTime();
				lease.actEnd = term.getEndTime();
			} catch (Exception e) {
			} finally {
				leasesLock.writeLock().unlock();
			}
		}
		
		@Override
		public void onTicket(IReservation r) {
			updateIfFailed(r);
		}
		
		@Override
		public void onTicketComplete(IReservation r) {
			Lease lease = leaseUpdateState(r, Lease.TICKETED);
			// Note the requested start and end times for inclusion
			// in the getLeases() status request.
			leasesLock.writeLock().lock();
			try {
				Term term = r.getRequestedTerm();
				lease.reqStart = term.getStartTime();
				lease.reqEnd = term.getEndTime();
			} catch (Exception e) {
			} finally {
				leasesLock.writeLock().unlock();
			}
			
		}
		
	}

	public static long ClockSkew = 1;
	public static final String PropertyMasterIP = "master.ipaddress";
	public static final String MasterConfig = "/scripts/dieselnet/dieselnet.xml";
	public static final String PropertyMasterName = "master.name";
	public static final String PropertyReservationType = "reservationType";
	public static final String PropertyMaster = "master";
	protected ImageDescriptor image;
	protected IServiceManager sm;
	protected ResourceType rtype = new ResourceType(1);
	private boolean initialized = false;
	protected LoggingTool logger = null;
	protected boolean needAdvanceTime = true;
	protected boolean needReservation = true;
	protected long advanceTime = 0;
	protected ISlice slice;
	protected String root;
	protected String dieselnetRoot;
	protected boolean controllerEnabled = true;
	
	private WebServer webServer = null;
	
	// The following data structures are used to keep track of
	// DieselNet leases that have passed through this system. This
	// information is for the DieselNet portal (or others) to get
	// status. See the getLeases() method.
	private class Lease {
		static final int REQUESTED = 0;
		static final int NASCENT = 1;
		static final int TICKETING = 2;
		static final int TICKETED = 3;
		static final int LEASING = 4;
		static final int LEASED = 5;
		static final int CLOSING = 6;
		static final int CLOSED = 7;
		static final int FAILED = 8;
		int state;
		String GID;
		String instance;
		String user;
		int duration;
		long updated;
		Date reqStart = null;
		Date reqEnd = null;
		Date actStart = null;
		Date actEnd = null;
		Lease(String GID, String instance, String user, int duration) {
			this.GID = GID;
			this.instance = instance;
			this.user = user;
			this.duration = duration;
			this.state = REQUESTED;
			this.updated = System.currentTimeMillis();
		}
	}
	private HashMap <String,Lease>leases = new HashMap<String,Lease>();
	private ReentrantReadWriteLock leasesLock = new ReentrantReadWriteLock();
	
	/* For initiating lease requests to ORCA. An XML-RPC lease request is
	 * received by a method in the DieselnetHandlers object. The method
	 * queues a LeaseRequest to requestLeaseQ. The request is picked
	 * up by this object in tick().
	 */
	private class LeaseRequest {
		int duration;
		long delay;
		String experiment;
		String instance;
		String user;
		LeaseRequest(int dur, long delay, String exp, String instance, String user) {
			this.duration = dur;
			this.delay = delay;
			this.experiment = exp;
			this.instance = instance;
			this.user = user;
		}
	}
	private LinkedList <LeaseRequest>requestLeaseQ = new LinkedList<LeaseRequest>();
	private int sizeLeaseQ = 0;  // modify only with requestLeaseQ synchronization
	
	/* A thread garbage collects non-active entries in the requestLeaseQ
	 * hash. Below we define what constitutes a stale entry, how often
	 * to perform garbage collection. 
	 */
	//private static final long gcSleepTime = 1000 * 60;  // testing
	//private static final long gcCollectTime = 1000 * 60 * 3;  // testing
	private static final long gcSleepTime = 1000 * 60 * 10;        // 10 minutes
	private static final long gcCollectTime = 1000 * 60 * 60 * 2;  // 2 hours
	private Thread gcThread = null;
	private static final String GCTHREAD = "dieselnet-gc";
	
	/*
	 * The following are property keys for values that we attach to
	 * tickets. These values can then be used for DieselNet-specific
	 * processing as we deal with with leases.
	 */
	private static final String  keyExperiment = "domeExperiment";
	private static final String  keyInstance = "domeInstance";
	private static final String keyUser = "domeUser";
	private static final String keyDuration = "domeDuration";
	
	public DieselnetController() {
		super();
		webServer = new WebServer(DieselnetConstants.WebServerPort);
		if (DieselnetConstants.DEBUGGING) {
			System.out.println("DieselnetController object instantiated");
		}
	}

	public synchronized void disableController() {
		this.controllerEnabled = false;
	}

	public synchronized void enableController() {
		this.controllerEnabled = true;
	}

	public IActor getActor() {
		return sm;
	}

	/**
	 * Calls up to a default broker to query and set the advance time.
	 */
	protected long getAdvanceTime() {
		long result = 0;
		Properties myProps = new Properties();
		myProps.setProperty("advanceTime", "null");

		try {
			// Properties returnProps = sm.getDefaultBroker().query(myProps, sm.getIdentity());
			// result = Long.parseLong(returnProps.getProperty("advanceTime"));
			result = 3;
		} catch (Exception e) {
			logger.exception("Agent does not support advanceTime ", e);
		}

		return result + ClockSkew;
	}

	public LoggingTool getLogger() {
		return logger;
	}

	private ICodClientReservation getMaster() {
		return getReservation(PropertyMaster);
	}

	public Node getMasterNode() {
		Node masterNode = null;

		ICodClientReservation master = getMaster();

		if ((master != null) && master.isActive()) {
			ResourceSet set = master.getLeasedResources();

			if (set != null) {
				NodeGroup group = (NodeGroup) set.getResources();

				if (group != null) {
					Node[] nodes = group.getNodesLocked();

					if ((nodes != null) && (nodes.length > 0)) {
						masterNode = nodes[0];
					}
				}
			}
		}

		return masterNode;
	}
	
	/**
	 * This is the method for creating the reservation request.
	 * 
	 * @param cycle - The current cycle time.
	 * @param lrq - A LeaseRequest object containing information about what
	 *    has been requested. 
	 * @return - The reservation request object.
	 */
	private ICodClientReservation getNewMaster(long cycle, LeaseRequest lrq) {
		ResourceSet rset = new ResourceSet(1, rtype);
		Properties request = rset.getRequestProperties();
		
		// Create a Date object for the requested start time. We start by
		// creating an object for the current cycle + advanceTime, then we
		// add in the user-requested delay.
		Date cycleDate = sm.getActorClock().date(cycle + advanceTime);
		Term term = new Term(new Date(cycleDate.getTime() + lrq.delay), lrq.duration);

		ICodClientReservation master;
		master = (ICodClientReservation) CodServiceManagerReservationFactory
				.getInstance().create(rset, term, slice, sm.getDefaultBroker());

		// These are meaningless in the DieselNet context, but we set them as
		// if we are requesting 100% of the resource pool. We behave as though
		// we are leasing Xen VMs.
		ResourceProperties.setMax(request, ResourceProperties.PropertyCpu, 100);
		ResourceProperties.setMin(request, ResourceProperties.PropertyCpu, 100);
		ResourceProperties.setMax(request, ResourceProperties.PropertyMemory, 128);
		ResourceProperties.setMin(request, ResourceProperties.PropertyMemory, 128);
		ResourceProperties.setMax(request, ResourceProperties.PropertyBandwidth, 100);
		ResourceProperties.setMin(request, ResourceProperties.PropertyBandwidth, 100);
		ResourceProperties.setMax(request, ResourceProperties.PropertyStorage, 0);
		ResourceProperties.setMin(request, ResourceProperties.PropertyStorage, 0);
		ResourceProperties.setMax(request, ResourceProperties.PropertyUnits, 1);
		ResourceProperties.setMin(request, ResourceProperties.PropertyUnits, 1);

		PropList.setProperty(request, BrokerSimplePolicy.REQUEST_TYPE, sm.getName());

		master.setLocalProperty(PropertyReservationType, PropertyMaster);

		// Set the DieselNet-specific properties. They will then be
		// passed through ORCA as opaque properties.
		rset.getConfigurationProperties().setProperty(keyExperiment, lrq.experiment);
		rset.getConfigurationProperties().setProperty(keyInstance, lrq.instance);
		rset.getConfigurationProperties().setProperty(keyUser, lrq.user);
		rset.getConfigurationProperties().setProperty(keyDuration, Integer.toString(lrq.duration));
		
		master.setServiceConfig(root + MasterConfig);
		master.setRenewable(false); // Set to FALSE (i.e. expires)
		//master.markVisible();  // ignore -- for public IP address
		PropertiesManager.setElasticSize(rset, false);
		// Set to true to allow ORCA to shift schedule if requested time
		// slot is not available.
		PropertiesManager.setElasticTime(rset, true);

		return master;
	}

	private ICodClientReservation getReservation(String s) {
		IReservation[] res = (IReservation[]) sm.getReservations(slice
				.getSliceID());

		for (int i = 0; i < res.length; i++) {
			IReservation reservation = res[i];

			String type = getReservationType(reservation);

			if ((type != null) && type.equals(s)) {
				return (ICodClientReservation) res[i];
			}
		}
		
		return null;
	}

	protected String getReservationType(IReservation reservation) {
		ResourceSet rset = reservation.getResources();

		if (rset != null) {
			Properties p = rset.getLocalProperties();

			if (p != null) {
				return p.getProperty(PropertyReservationType);
			}
		}

		return null;
	}

	public ISlice getSlice() {
		return slice;
	}

	/**
	 * Initialization called soon after the object is instantiated. Among
	 * other things, it creates the XML-RPC web server and starts a thread
	 * to listen for procedure calls.
	 */
	public synchronized void initialize() throws Exception {
		if (!initialized) {
			if (sm == null) {
				throw new Exception("Missing actor");
			}

			if (slice == null) {
				throw new Exception("Missing slice");
			}

			logger = sm.getLogger();
			root = ((Manager) ContainerManager.getInstance()).getPackageRootFolder(DieselnetConstants.PackageId);
			((ServiceManagerCalendarPolicy) sm.getPolicy()).register(new DieselnetEventHandler());
			
			image = new ImageDescriptor();
			// For now we only attach the guid. Assume the image is present.
			// Not really valid in the DieselNet context.
			image.setGuid(DieselnetConstants.ImageGuid);
			
			// Initialize the XML-RPC web server so that we can receive messages
			// from the DieselNet portal (or from elsewhere).
			try {
				XmlRpcServer rpcServer = webServer.getXmlRpcServer();
				PropertyHandlerMapping phm = new PropertyHandlerMapping();
				// Initialize static variables in the handlers object.
				DieselnetHandlers dnh = new DieselnetHandlers(this, logger);
				phm.addHandler("DieselnetHandlers", dnh.getClass());
				rpcServer.setHandlerMapping(phm);
				XmlRpcServerConfigImpl serverConfig = (XmlRpcServerConfigImpl) rpcServer.getConfig();
				serverConfig.setEnabledForExceptions(true);
				serverConfig.setContentLengthOptional(false);
				// All set ...
				webServer.start();
			} catch (Exception e) {
				if (logger != null) {
					logger.critical("Exception creating DieselNet web server: " + e);
				}
				throw new Exception("DieselNet WebServer creation error");
			}
			
			// Start a garbage collection thread to clean up stale lease
			// information.
			gcThread = new Thread(this, GCTHREAD);
	        gcThread.start();
						
			initialized = true;
		}
	}
	
	/**
	 * Starts threads.
	 */
	public void run() {
        if (Thread.currentThread().getName().equals(GCTHREAD)) {
        	gcThread();
        }
	}
	
	/**
	 * Garbage collection thread. Cleans up information on leases that
	 * have closed/failed in the past. We don't want the hash to grow
	 * boundlessly.
	 */
	private void gcThread() {
		while (true) {
			try {
				Thread.sleep(gcSleepTime);
				gcLeases();
			} catch (Exception e) {
			}
		}
	}

	/*
	 * Called by DieselnetManagerObject. (It's purpose is unknown ...)
	 */
	public synchronized boolean isControllerEnabled() {
		return this.controllerEnabled;
	}

	/*
	 * (non-Javadoc)
	 * @see orca.shirako.api.ISerialize#reset(java.util.Properties)
	 */
	public void reset(Properties p) throws Exception {
	}

	/*
	 * Persistence and recovery code. XXX: the current implementation is not
	 * persistent and cannot recover!!!
	 */
	public Properties save() throws Exception {
		Properties p = new Properties();
		save(p);

		return p;
	}

	/*
	 * (non-Javadoc)
	 * @see orca.shirako.api.ISerialize#save(java.util.Properties)
	 */
	public void save(Properties p) throws Exception {
	}


	/*
	 * (non-Javadoc)
	 * @see orca.shirako.api.ILoadSource#setActor(orca.shirako.api.IActor)
	 */
	public void setActor(IActor sm) {
		this.sm = (IServiceManager) sm;
	}

	/*
	 * (non-Javadoc)
	 * @see orca.shirako.api.IController#setSlice(orca.shirako.api.ISlice)
	 */
	public void setSlice(ISlice slice) {
		this.slice = slice;
	}

	/**
	 * Informs the load source that another cycle has passed. Initiates the
	 * formation of demand for the service manager.
	 * 
	 * @param cycle - This cycle
	 */
	public void tick(long cycle) {
		if (needAdvanceTime) {
			advanceTime = getAdvanceTime();
			needAdvanceTime = false;
		}
		// Determine if we have a new lease request. If we do, generate a
		// a request to ORCA for a ticket. We maintain the sizeLeaseQ variable
		// to avoid synchronizing on the lease request queue every tick.
		if (sizeLeaseQ > 0) {
			LeaseRequest lrq = dequeueLeaseRequest();
			if (lrq != null) {
				if (leaseAdd(lrq) == null) {
					logger.info("Ignoring duplicate DieselNet lease request");
				} else {
					logger.debug("Requesting a DieselNet lease");
					synchronized (this) {
						sm.demand(getNewMaster(cycle, lrq));
					}
				}
			}
		}
	}
		
	/**
	 * Queues a request for a lease. This method is called by the
	 * DieselnetHandlers object when a request lease RPC is received.
	 *  
	 * @param duration - Requested duration of the lease, in milliseconds.
	 * @param delay - Requested delay before starting the lease, in millis,
	 *    relative to whenever the request is eventually processed. 
	 * @param experiment - The Global Experiment Id identifying the
	 *    experiment requesting the lease.
	 * @param instance - The instance identifier to be associated with
	 *    the lease.
	 * @param userId - DOME user associated with the request.
	 */
	public void queueLeaseRequest(int duration, long delay,
			String experiment, String instance, String userId) {
		LeaseRequest lr =
			new LeaseRequest(duration, delay, experiment, instance, userId);
		synchronized (requestLeaseQ) {
			sizeLeaseQ++;
			requestLeaseQ.add(lr);
		}
	}
	
	/**
	 * Dequeues a lease request. It's best to check sizeLeaseQ for a
	 * non-zero value rather than call this method every tick.
	 * 
	 * @return A LeaseRequest object, or null.
	 */
	private LeaseRequest dequeueLeaseRequest() {
		synchronized (requestLeaseQ) {
			LeaseRequest lr = requestLeaseQ.poll();
			if (lr == null) {
				return null;
			}
			sizeLeaseQ--;
			return lr; 
		}
	}
	
	/**
	 * Adds a new Lease to the hash of known leases.
	 * 
	 *  @param GID - Experiment global identifier.
	 *  @param instance - Instance of the experiment.
	 *  @param user - DOME user id associated with the experiment.
	 *  @param duration - Requested experiment duration, ins milliseconds.
	 *  
	 *  @return Returns the Lease object if added, else null. Null will be
	 *     be returned if an entry for the key already existed.
	 */
	private Lease leaseAdd(String GID, String instance, String user, int duration) {
		leasesLock.writeLock().lock();
		try {
			String key = leaseHashKey(GID, instance);
			if (leases.containsKey(key)) {
				return null;
			}
			Lease lease = new Lease(GID, instance, user, duration);
			leases.put(key, lease);
			return lease;
		} finally {
			leasesLock.writeLock().unlock();
		}
	}
	
	/**
	 * Overload of leaseAdd. Takes the input from a LeaseRequest object.
	 * 
	 * @param lrq - The LeaseRequest object.
	 * @return True if an entry was added, else false.
	 */
	private Lease leaseAdd(LeaseRequest lrq) {
		return leaseAdd(lrq.experiment, lrq.instance, lrq.user, lrq.duration);
	}
	
	/**
	 * Updates the state in a Lease object.
	 * 
	 * @param r - Reservation object given to us by ORCA.
	 * @param state - New state.
	 * @return The Lease object.
	 */
	private Lease leaseUpdateState(IReservation r, int state) {
		String expGID = r.getResources().getConfigurationProperties().getProperty(keyExperiment);
		String instance = r.getResources().getConfigurationProperties().getProperty(keyInstance);
		if ((expGID == null) || (instance == null)) {
			// Paranoia.
			return null;
		}
		// We should have knowledge of the lease. If we don't, e.g. this
		// controller had been restarted, then add a lease entry. The lock
		// is reentrant so we can lock it again in leaseAdd().
		String key = leaseHashKey(expGID, instance);
		leasesLock.writeLock().lock();
		try {
			Lease lease = leases.get(key);
			if (lease == null) {
				// We need to create an entry.
				String user = r.getResources().getConfigurationProperties().getProperty(keyUser);
				String dur = r.getResources().getConfigurationProperties().getProperty(keyDuration);
				lease = leaseAdd(expGID, instance, user, new Integer(dur));
			}
			// Update the state and time. If the reservation indicates that
			// it has failed, ignore the state parameter and set the state
			// to FAILED.
			lease.state = (r.isFailed()) ? Lease.FAILED : state;
			lease.updated = System.currentTimeMillis();
			return lease;
		} catch (Exception e) {
			logger.error("Exception updating DieselNet lease object: " + e);
			return null;
		} finally {
			leasesLock.writeLock().unlock();
		}
	}
	
	/**
	 * Creates the hash key for the leases hash. This is a concatenation
	 * of the instance and global experiment id. Together, they should be
	 * unique. (Actually, an instance id should be unique, but we don't
	 * rely on it.)
	 * 
	 * @param GID - Experiment global identifier.
	 * @param instance - Instance of the experiment.
	 * @return The key.
	 */
	private String leaseHashKey(String GID, String instance) {
		return instance + ":" + GID;
	}
	
	/**
	 * This method does the work for the XML-RPC getLeases() handler. It
	 * returns information for each known lease.
	 * 
	 * @param user - If not null, then limit the output to leases associated
	 *    with the specified user.
	 * @return XML-like data describing the leases.
	 */
	public String getLeases(String user) {
		leasesLock.readLock().lock();
		try {
			String out = "";
			// For each known lease, generate output.
			Iterator <Lease> it = leases.values().iterator();
			while (it.hasNext()) {
				Lease lease = it.next();
				// Skip if not for a specific user?
				if (user != null) {
					if (user.equals(lease.user)) {
						continue;
					}
				}
				out += "<lease>";
				out += "<experiment>" + lease.GID + "</experiment>";
				out += "<instance>" + lease.instance + "</instance>";
				out += "<user>" + lease.user + "</user>";
				out += "<duration>" + lease.duration + "</duration>";
				out += "<state value=\"" + lease.state + "\">";
				out += strState(lease.state);
				out += "</state>";
				if (lease.state >= Lease.TICKETING) {
					// Format the requested lease times from the reservation.
					if ((lease.reqStart != null) && (lease.reqEnd != null)) {
						try {
							String req = "<requested>";
							req += "<start>" + lease.reqStart.getTime() + "</start>";
							req += "<end>" + lease.reqEnd.getTime() + "</end>";
							out += req + "</requested>";
						} catch (Exception ex) {
						}
					}
				}
				if (lease.state >= Lease.LEASED) {
					// Format the actual lease times from the reservation.
					if ((lease.actStart != null) && (lease.actEnd != null)) {
						try {
							String act = "<actual>";
							act += "<start>" + lease.actStart.getTime() + "</start>";
							act += "<end>" + lease.actEnd.getTime() + "</end>";
							out += act + "</actual>";
						} catch (Exception ex) {
						}
					}
				}
				out += "</lease>";
			}
			return "<leases>" + out + "</leases>"; 
		} catch (Exception e) {
			logger.error("Exception iterating through DieselNet leases: " + e);
			return "<leases></leases>";
		} finally {
			leasesLock.readLock().unlock();
		}
	}
	
	/**
	 * Given a lease state value, returns a string describing the state.
	 * 
	 * @param state - State value.
	 * @return - Description of the state.
	 */
	private String strState(int state) {
		switch (state) {
		case Lease.REQUESTED:
			return "requested";
		case Lease.NASCENT:
			return "nascent";
		case Lease.TICKETING:
			return "ticketing";
		case Lease.TICKETED:
			return "ticketed";
		case Lease.LEASING:
			return "leasing";
		case Lease.LEASED:
			return "leased";
		case Lease.CLOSING:
			return "closing";
		case Lease.CLOSED:
			return "closed";
		case Lease.FAILED:
			return "failed";
		default:
			return "unknown";
		}
	}
	
	/**
	 * Garbage collects leases in the hash map.
	 */
	private void gcLeases() {
		LinkedList <Lease>expired = new LinkedList<Lease>();
		// We go through the list of leases with the read lock to determine
		// which are old enough to garbage collect. Any lease that may be
		// removed is placed on the expired list.
		long now = System.currentTimeMillis();
		leasesLock.readLock().lock();
		try {
			Iterator <Lease> it = leases.values().iterator();
			while (it.hasNext()) {
				Lease lease = it.next();
				if (lease.state < Lease.CLOSED) {
					continue;
				}
				if ((now - lease.updated) < gcCollectTime) {
					continue;
				}
				expired.add(lease);
			}
		} catch (Exception e) {
		} finally {
			leasesLock.readLock().unlock();
		}
		// Do we have anything to garbage collect?
		if (expired.size() == 0) {
			return;
		}
		// Now get the write lock and remove entries from the hash.
		leasesLock.writeLock().lock();
		try {
			Iterator <Lease> it = expired.iterator();
			while (it.hasNext()) {
				Lease ls = it.next();
				leases.remove(leaseHashKey(ls.GID, ls.instance));
			}
		} finally {
			leasesLock.writeLock().unlock();
		}
	}
}