/*
 * 
 ************************************************************************
 *
 * Copyright (c) 2009 University of Massachusetts, Amherst
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and/or hardware specification (the Work) to
 * deal in the Work without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Work, and to permit persons to
 * whom the Work is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Work.
 *
 * THE WORK IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE WORK OR THE USE OR OTHER DEALINGS
 * IN THE WORK.
 *
 ************************************************************************
 *
 * This object contains methods that are invoked via external XML-RPC calls.
 */

package orca.controllers.dieselnet;

import orca.util.LoggingTool;

public class DieselnetHandlers {
	
	//private static final int minDurMins = 24 * 60;         // 1 day
	private static final int minDurMins = 1;  // XXX: temporary value
	private static final int maxDurMins = 8 * 24 * 60;     // 8 days
	private final int minDelayMins = 0;
	private final int maxDelayMins = 20 * 24 * 60;         // 20 days
	
	private static LoggingTool logger = null;
	private static DieselnetController controller = null;
	
	
	/**
	 * Constructor that gets called when invoking XML-RPC methods.
	 */
	public DieselnetHandlers() {
		
	}
	
	/**
	 * Special constructor used by the Controller before configuring
	 * XML-RPC calls. Its purpose is to set static values, most notably
	 * the Controller's object so that this class may invoke methods in
	 * the Controller.
	 * 
	 * @param controller - The instantiated DieselNet Controller object. There
	 *    should be only one running.
	 * @param logger - Handle to the logger.
	 */
	public DieselnetHandlers(DieselnetController controller, LoggingTool logger) {
		DieselnetHandlers.controller = controller;
		DieselnetHandlers.logger = logger;
	}
	
	/**
	 * Handles requests for experiment leases. After performing some simple
	 * validation of the parameters, queue a request for a lease to the
	 * DieselnetController. 
	 * 
	 * @param expGID - Unique Global Experiment Identifier representing a
	 *    DOME experiment.
	 * @param instance - Experiment instance.
	 * @param userId - DOME user associated with the request.
	 * @param mnDuration - Requested duration of the experiment, in minutes.
	 * @param mnDelay - Request delay for the experiment to start, in minutes.
	 * @return 0 if a request was queued to the controller, else a non-zero
	 *    value.
	 */
	public int requestLease(String expGID, int instance, String userId,
			int mnDuration, int mnDelay) {
		if ((expGID.length() == 0) || (userId.length() == 0)) {
			return 1;
		}
		// Make certain values are within range.
		if (mnDuration < minDurMins) {
			mnDuration = minDurMins;
		} else if (mnDuration > maxDurMins) {
			mnDuration = maxDurMins;
		}
		if (mnDelay < minDelayMins) {
			mnDelay = minDelayMins;
		} else if (mnDelay > maxDelayMins) {
			mnDelay = maxDelayMins;
		}
		if (logger != null) {
			logger.debug(
				"DieselNet lease request: " + expGID +
				":" + instance + " (" + mnDuration + "," + mnDelay + ")"
				);
		}
		// Convert minutes to milliseconds and have the Controller queue
		// the request.
		int duration = mnDuration * 60 * 1000;
		long delay = mnDelay * 60 * 1000;
		controller.queueLeaseRequest(duration, delay, expGID,
				Integer.toString(instance), userId);
		return 0;
	}
	
	/**
	 * This method requests a list of leases known by the controller.
	 * 
	 * @param userId - If not an empty string then limit the output to
	 *    leases associated with the specified DOME user id.
	 * @return An XML-like string representing the leases.
	 */
	public String getLeases(String userId) {
		String user = null;
		if (userId.length() > 0) {
			user = userId;
		}
		return controller.getLeases(user);
	}
}
