/*
 ************************************************************************
 *
 * Copyright (c) 2009 University of Massachusetts, Amherst
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and/or hardware specification (the "Work") to
 * deal in the Work without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Work, and to permit persons to
 * whom the Work is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Work.
 *
 * THE WORK IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE WORK OR THE USE OR OTHER DEALINGS
 * IN THE WORK.
 *
 ************************************************************************
 *
 * Abstract class used by the DieselNet handler. This class extends
 * the Apache Ant Task class.
 */

package orca.handlers.dieselnet;

import org.apache.tools.ant.Task;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;

public abstract class DieselnetTask extends Task {
	
	private static final String defaultHost = "127.0.0.1";
	private static final int defaultPort = 9237;
	
	private static final String exitCodePropertyName = "dieselnet.exit";
	
	// Exit codes. We always return 0 since a non-zero value will likely
	// make ORCA think that the problem is not transient but associated
	// with the resource, i.e. we'll stop getting leases.
	protected static final int EXITUNEXPECTED = 0;
	protected static final int EXITRPCCLIENT = 0;
	protected static final int EXITRPCFAILED = 0;
	
	protected String domeServer = "http://" + defaultHost + ":" + defaultPort;
	protected String domeExperiment = null;
	protected String domeInstance = null;
	protected String domeUser = null;
	protected String domeTicketId = null;
	private String exitCodeProperty = exitCodePropertyName;
	
	private Project project = null;
	
	/**
	 * Where the class extending this class begins life.
	 * 
	 * @return Result code to be returned by the ant task.
	 */
	abstract protected int doExecute() throws BuildException;
	
	/**
	 * This is the method that is invoked after the Task is instantiated
	 * and properties have been set.
	 */
	public void execute() throws BuildException {
		// Get the project object.
		project = getProject();
		try {
			// Verify required properties have been defined.
			if ((domeExperiment == null) || (domeInstance == null) || (domeServer == null)) {
				throw new BuildException("Property required by DOME is not set ");
			}
			// Call the subclass to do its thing. Set the tasks exit code.
			int val = doExecute();
			if ((project != null) && (exitCodeProperty != null)) {
				project.setProperty(exitCodeProperty, Integer.toString(val));
			}
		} catch (Exception e) {
			if ((project != null) && (exitCodeProperty != null)) {
				project.setProperty(exitCodeProperty, Integer.toString(EXITUNEXPECTED));
			}
			throw new BuildException(e);
		}
	}
	
	/*
	 * The following setXXXX methods are invoked by ant to set property values.
	 * For example, <DieselneTask paramName="${script.value}"/> would invoke
	 * method setParamName() with the value that is assigned to script.name 
	 * within the ant script. Once all properties have been set, control is
	 * passed to us by ant invoking the above execute() method.
	 */
	
	/**
	 * Defines the name of the property used to store the exit code.
	 * 
	 * @param exitCodeProperty - Property name.
	 */
	public void setExitCodeProperty(String exitCodeProperty) {
		this.exitCodeProperty = exitCodeProperty;
	}
	
	/**
	 * Returns the name of the exit code property string.
	 * 
	 * @return Property name.
	 */
	public String getExitCodeProperty() {
		return this.exitCodeProperty;
	}
	
	/**
	 * Sets the domeServer property indicating the server URL to send
	 * XML-RPC requests to.
	 * 
	 * @param domeServer - Node / address of the server.
	 */
	public void setDomeServer(String domeServer) {
		this.domeServer = domeServer;
	}
	
	/**
	 * Gets the domeServer URL property value.
	 * 
	 * @return Node / address of the server.
	 */
	public String getDomeServer() {
		return this.domeServer;
	}
	
	/**
	 * Sets the experiment global identifier (GID).
	 * 
	 * @param domeExperiment
	 */
	public void setDomeExperiment(String domeExperiment) {
		this.domeExperiment = domeExperiment;
	}
	
	/**
	 * Gets the domeExperiment property value.
	 * 
	 * @return The experiment GID.
	 */
	public String getDomeExperiment() {
		return this.domeExperiment;
	}
	
	/**
	 * Sets the experiment instance.
	 * 
	 * @param domeInstance - Instance value.
	 */
	public void setDomeInstance(String domeInstance) {
		this.domeInstance = domeInstance;
	}
	
	/**
	 * Gets the domeInstance property value.
	 * 
	 * @return The experiment instance identifier.
	 */
	public String getDomeInstance() {
		return this.domeInstance;
	}
	
	/**
	 * Sets the user id.
	 * 
	 * @param domeUser - The user.
	 */
	public void setDomeUser(String domeUser) {
		this.domeUser = domeUser;
	}
	
	/**
	 * Gets the domeUser property value.
	 * 
	 * @return The user id.
	 */
	public String getDomeUser() {
		return this.domeUser;
	}
	
	/**
	 * Sets the ticket id, an opaque value defined when the ticket is
	 * requested by the portal.
	 * 
	 * @param domeTicketId - The ticket identifier string.
	 */
	public void setDomeTicketId(String domeTicketId) {
		this.domeTicketId = domeTicketId;
	}
	
	/**
	 * Returns the DOME-defined ticket id.
	 * 
	 * @return The ticket id.
	 */
	public String getDomeTicketId() {
		return this.domeTicketId;
	}

}
